// Package openapi3 parses and writes OpenAPI 3 specifications.
//
// The OpenAPI 3.0 specification can be found at:
//   https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.md
package openapi3
